@@ -1,21 +1,15 @@
--- tkz_elements_functions_lines.lua
--- date 2025/06/15
--- version 4.15c
--- Copyright 2025  Alain Matthes
--- This work may be distributed and/or modified under the
--- conditions of the LaTeX Project Public License, either version 1.3
--- of this license or (at your option) any later version.
--- The latest version of this license is in
--- http://www.latex-project.org/lppl.txt
--- and version 1.3 or later is part of all distributions of LaTeX
--- version 2005/12/01 or later.
--- This work has the LPPL maintenance status “maintained”.
--- The Current Maintainer of this work is Alain Matthes.
+-- File: tkz_elements_functions_lines.lua
+-- Copyright (c) 2023–2025 Alain Matthes
+-- SPDX-License-Identifier: LPPL-1.3c
+-- Maintainer: Alain Matthes
 
 function length_(a, b)
-  return point.abs(a - b)
+	return point.abs(a - b)
 end
 
+function distance_(L, pt)
+	return point.mod(projection_(L.pa, L.pb, pt) - pt)
+end
 ---------------------------------------------------------------------------
 --                 Lines
 ---------------------------------------------------------------------------
@@ -175,32 +169,32 @@ function orthogonal_at_(za, zb, pt, k)
 end
 
 function bisector(a, b, c)
-  local i = in_center_(a, b, c)
-  return line:new(a, intersection_ll_(a, i, b, c))
+	local i = in_center_(a, b, c)
+	return line:new(a, intersection_ll_(a, i, b, c))
 end
 
 function altitude(a, b, c)
-  local o, p
-  -- Get the orthocenter (which is the point of concurrency of the altitudes)
-  o = ortho_center_(a, b, c)
+	local o, p
+	-- Get the orthocenter (which is the point of concurrency of the altitudes)
+	o = ortho_center_(a, b, c)
 
-  -- Get the perpendicular projection of point 'a' onto the line defined by 'b' and 'c'
-  p = projection_(b, c, a)
+	-- Get the perpendicular projection of point 'a' onto the line defined by 'b' and 'c'
+	p = projection_(b, c, a)
 
-  -- Return the altitude, which is the line from point 'a' to the point 'p'
-  return line:new(a, p)
+	-- Return the altitude, which is the line from point 'a' to the point 'p'
+	return line:new(a, p)
 end
 
 function bisector_ext(a, b, c) -- n=1 swap n=2 swap 2
-  local i, p
-  -- Get the incenter of the triangle
-  i = in_center_(a, b, c)
+	local i, p
+	-- Get the incenter of the triangle
+	i = in_center_(a, b, c)
 
-  -- Rotate the incenter by 90 degrees around point 'a' to compute the external bisector
-  p = rotation_(a, math.pi / 2, i)
+	-- Rotate the incenter by 90 degrees around point 'a' to compute the external bisector
+	p = rotation_(a, math.pi / 2, i)
 
-  -- Return the external bisector as the line passing through points 'a' and 'p'
-  return line:new(a, p)
+	-- Return the external bisector as the line passing through points 'a' and 'p'
+	return line:new(a, p)
 end
 -- orthonormal cartesian coordinate system
 -- function occs_(p,za,zb)
