@@ -1,75 +1,73 @@
--- tkz_elements_point.lua
--- date 2025/06/15
--- version 4.15c
--- Copyright 2025  Alain Matthes
--- This work may be distributed and/or modified under the
--- conditions of the LaTeX Project Public License, either version 1.3
--- of this license or (at your option) any later version.
--- The latest version of this license is in
--- http://www.latex-project.org/lppl.txt
--- and version 1.3 or later is part of all distributions of LaTeX
--- version 2005/12/01 or later.
--- This work has the LPPL maintenance status “maintained”.
--- The Current Maintainer of this work is Alain Matthes.
+-- File: tkz_elements_point.lua
+-- Copyright (c) 2023–2025 Alain Matthes
+-- SPDX-License-Identifier: LPPL-1.3c
+-- Maintainer: Alain Matthes
 
 -- point.lua
 -- Définition locale de class
+
 local function class(init, base)
-  local cls = {}
-  cls.__index = cls
+	local cls = {}
+	cls.__index = cls
 
-  if base then
-    setmetatable(cls, { __index = base })
-    cls._base = base
-  end
+	if base then
+		setmetatable(cls, { __index = base })
+		cls._base = base
+	end
 
-  setmetatable(cls, {
-    __call = function(class_tbl, ...)
-      local self = setmetatable({}, class_tbl)
-      if init then
-        init(self, ...)
-      elseif base and base.init then
-        base.init(self, ...)
-      end
-      return self
-    end
-  })
+	setmetatable(cls, {
+		__call = function(class_tbl, ...)
+			local self = setmetatable({}, class_tbl)
+			if init then
+				init(self, ...)
+			elseif base and base.init then
+				base.init(self, ...)
+			end
+			return self
+		end,
+	})
 
-  cls.init = init
+	cls.init = init
 
-  function cls:is_a(klass)
-    local m = getmetatable(self)
-    while m do
-      if m == klass then return true end
-      m = m._base
-    end
-    return false
-  end
+	function cls:is_a(klass)
+		local m = getmetatable(self)
+		while m do
+			if m == klass then
+				return true
+			end
+			m = m._base
+		end
+		return false
+	end
 
-  return cls
+	return cls
 end
 
 point = class(function(p, re, im)
-  -- Cas 1 : appel standard point(x, y)
-  if type(re) == "number" and type(im) == "number" then
-    p.re = re
-    p.im = im
+	-- Cas 1 : appel standard point(x, y)
+	if type(re) == "number" and type(im) == "number" then
+		p.re = re
+		p.im = im
 
-  -- Cas 2 : appel interne point({re=..., im=...})
-  elseif type(re) == "table" and im == nil and type(re.re) == "number" and type(re.im) == "number" then
-    p.re = re.re
-    p.im = re.im
+	-- Cas 2 : appel interne point({re=..., im=...})
+	elseif type(re) == "table" and im == nil and type(re.re) == "number" and type(re.im) == "number" then
+		p.re = re.re
+		p.im = re.im
 
-  -- Cas invalide
-  else
-    error("Invalid point: expected two numbers or a complex {re,im}, got re = "
-      .. tostring(re) .. ", im = " .. tostring(im))
-  end
+	-- Cas invalide
+	else
+		tex.error(
+			"Invalid point: expected two numbers or a complex {re,im}, got re = "
+				.. tostring(re)
+				.. ", im = "
+				.. tostring(im)
+		)
+	end
 
-  p.type = "point"
-  p.argument = math.atan(p.im, p.re)
-  p.modulus  = math.sqrt(p.re * p.re + p.im * p.im)
-  p.mtx      = matrix:new({ { p.re }, { p.im } })
+	p.type = "point"
+	p.argument = math.atan(p.im, p.re)
+	p.modulus = math.sqrt(p.re * p.re + p.im * p.im)
+	p.mtx = matrix:new({ { p.re }, { p.im } })
 end)
 
 local sqrt = math.sqrt
@@ -81,22 +79,47 @@ local min = math.min
 local max = math.max
 local abs = math.abs
 
-local function topoint(z1)
-	if type(z1) == "number" then
-		return point(z1, 0)
+-- local function topoint(z1)
+--   if type(z1) == "number" then
+--     return point(z1, 0)
+--   else
+--     return z1
+--   end
+-- end
+
+local function is_point(obj)
+	if type(obj) ~= "table" then
+		return false
+	end
+	local mt = getmetatable(obj)
+	while mt do
+		if mt == point then
+			return true
+		end
+		mt = mt._base
+	end
+	return false
+end
+
+local function topoint(z)
+	if is_point(z) then
+		return z
+	elseif type(z) == "number" then
+		return point(z, 0)
+	elseif type(z) == "table" and type(z.re) == "number" and type(z.im) == "number" then
+		return point(z) -- c'est géré dans le constructeur
 	else
-		return z1
+		tex.error("topoint: cannot convert value to point")
 	end
 end
 
+
 local function check(z1, z2)
-	if type(z1) == "number" then
-		return point(z1, 0), z2
-	elseif type(z2) == "number" then
-		return z1, point(z2, 0)
-	else
-		return z1, z2
+	local p1, p2 = topoint(z1), topoint(z2)
+	if not p1 or not p2 then
+		tex.error("check: expected point or number")
 	end
+	return p1, p2
 end
 -- -------------------------------------------------------------------
 -- metamethods
@@ -124,25 +147,25 @@ end
 
 -- dot product is '..'  (a+ib) (c-id) = ac+bd + i(bc-ad)
 function point.__concat(z1, z2)
-	local z
-	z = z1 * point.conj(z2)
+	local c1, c2 = check(z1, z2)
+	local z = c1 * point.conj(c2)
 	return z.re
 end
 
 -- determinant  is '^'   (a-ib) (c+id) = ac+bd + i(ad - bc)
 function point.__pow(z1, z2)
-	local z
-	z = point.conj(z1) * z2
+	local c1, c2 = check(z1, z2)
+	local z = point.conj(c1) * c2
 	return z.im
 end
 
-function point.__div(x, y)
-	local xx = topoint(x)
-	local yy = topoint(y)
-	return point(
-		(xx.re * yy.re + xx.im * yy.im) / (yy.re * yy.re + yy.im * yy.im),
-		(xx.im * yy.re - xx.re * yy.im) / (yy.re * yy.re + yy.im * yy.im)
-	)
+function point.__div(z1, z2)
+	local c1, c2 = check(z1, z2)
+	local den = c2.re * c2.re + c2.im * c2.im
+	if den == 0 then
+		tex.error("Division by zero in point.__div")
+	end
+	return point((c1.re * c2.re + c1.im * c2.im) / den, (c1.im * c2.re - c1.re * c2.im) / den)
 end
 
 function point.__tostring(z)
@@ -198,9 +221,8 @@ function point.__tonumber(z)
 end
 
 function point.__eq(z1, z2)
-  local epsilon = 1e-10 
-  return math.abs(z1.re - z2.re) < epsilon 
-      and math.abs(z1.im - z2.im) < epsilon
+	local epsilon = 1e-10
+	return math.abs(z1.re - z2.re) < epsilon and math.abs(z1.im - z2.im) < epsilon
 end
 -- -------------------------------------------------------------------
 local function pyth(a, b)
