@@ -1,16 +1,7 @@
--- tkz_elements_vectors.lua
--- date 2025/06/15
--- version 4.15c
--- Copyright 2025  Alain Matthes
--- This work may be distributed and/or modified under the
--- conditions of the LaTeX Project Public License, either version 1.3
--- of this license or (at your option) any later version.
--- The latest version of this license is in
--- http://www.latex-project.org/lppl.txt
--- and version 1.3 or later is part of all distributions of LaTeX
--- version 2005/12/01 or later.
--- This work has the LPPL maintenance status “maintained”.
--- The Current Maintainer of this work is Alain Matthes.
+-- File: tkz_elements_vectors.lua
+-- Copyright (c) 2023–2025 Alain Matthes
+-- SPDX-License-Identifier: LPPL-1.3c
+-- Maintainer: Alain Matthes
 
 -- ----------------------------------------------------------------------------
 vector = {}
@@ -22,21 +13,15 @@ function vector:new(za, zb)
 	local norm = point.mod(zb - za)
 	local mtx = matrix:new({ { za }, { zb } })
 	local z = zb - za
-	local vect = { tail = za, 
-                 head = zb, 
-                 z = z, 
-                 norm = norm, 
-                 mtx = mtx, 
-                 slope = slope, 
-                 type = type }
+	local vect = { tail = za, head = zb, z = z, norm = norm, mtx = mtx, slope = slope, type = type }
 	setmetatable(vect, self)
 	return vect
 end
 
 setmetatable(vector, {
-  __call = function(cls, ...)
-    return cls:new(...)
-  end
+	__call = function(cls, ...)
+		return cls:new(...)
+	end,
 })
 
 function vector.__add(v1, v2)
@@ -68,7 +53,7 @@ function vector.__concat(v1, v2)
 end
 
 function vector:get()
-  return self.pa, self.pb
+	return self.pa, self.pb
 end
 
 -- Normalize the vector (unit vector)
