@@ -13,8 +13,8 @@
 
 luatexbase.provides_module {
   name        = 'luatexko',
-  date        = '2025/07/16',
-  version     = '4.1',
+  date        = '2025/09/26',
+  version     = '4.7',
   description = 'typesetting Korean with LuaTeX',
   author      = 'Dohyun Kim, Soojin Nam',
   license     = 'LPPL v1.3+',
@@ -85,6 +85,7 @@ local discid    = node.id"disc"
 local glueid    = node.id"glue"
 local glyphid   = node.id"glyph"
 local hlistid   = node.id"hlist"
+local ins_id    = node.id"ins"
 local kernid    = node.id"kern"
 local localparid = node.id"local_par"
 local mathid    = node.id"math"
@@ -100,6 +101,7 @@ local italcorr   = 3
 local lua_number = 100
 local lua_value  = 108
 local spaceskip  = 13
+local parfillskip = 15
 local indentbox  = 3
 local nohyphen = registernumber"l@nohyphenation" or -1 -- verbatim
 local langkor  = registernumber"koreanlanguage"  or 16383
@@ -113,6 +115,7 @@ local dotemphattr      = attributes.luatexkodotemphattr
 local rubyattr         = attributes.luatexkorubyattr
 local hangulbyhangulattr = attributes.luatexkohangulbyhangulattr
 local hanjabyhanjaattr   = attributes.luatexkohanjabyhanjaattr
+local inhibitglueattr  = attributes.luatexkoinhibitglueattr
 
 local unicodeattr = new_attribute"luatexkounicodeattr"
 
@@ -291,7 +294,7 @@ local fontoptions = {
 
   en_size = setmetatable( {}, { __index = function(t, fid)
     if fid then
-      val = (get_font_param(fid, "quad") or 655360)/2
+      local val = (get_font_param(fid, "quad") or 655360)/2
       t[fid] = val
       return val
     end
@@ -806,15 +809,21 @@ local allowbreak_false_nodes = {
   [ruleid]    = true,
   [discid]    = true,
   [glueid]    = true,
-  [penaltyid] = true,
+  [mathid]    = true,
+  [dirid]     = true,
+  [glyphid]   = true,
 }
 
 local function is_blocking_node (curr)
   local id, subtype = curr.id, curr.subtype
+  if id == hlistid then
+    if subtype == indentbox then return false end
+    if curr.next and curr.next.id == ins_id then return false end -- footnote
+  end
   return allowbreak_false_nodes[id] or id == kernid and subtype == userkern
 end
 
-local function hbox_char_font (box, init, glyfonly)
+local function hbox_char_font (box, init, deep)
   local mynext = init and getnext  or getprev
   local curr   = init and box.list or nodeslide(box.list)
   while curr do
@@ -825,8 +834,10 @@ local function hbox_char_font (box, init, glyfonly)
         return c, curr.font
       end
     elseif curr.list then
-      return hbox_char_font(curr, init, glyfonly)
-    elseif not glyfonly and is_blocking_node(curr) then
+      return hbox_char_font(curr, init, deep)
+    elseif deep then
+    elseif curr.id == glueid and curr.width == 0 and has_attribute(box, rubyattr) then -- ruby
+    elseif is_blocking_node(curr) then
       return
     end
     curr = mynext(curr)
@@ -854,16 +865,24 @@ local function goto_end_actualtext (curr)
 end
 
 local function insert_glue_before (head, curr, par, br, brb, classic, ict, dim, fid)
-  local pn = nodenew(penaltyid)
-  if not br then
-    pn.penalty = 10000
-  elseif type(brb) == "number" then
-    pn.penalty = brb
-  elseif par and nodecount(glyphid, curr) <= 2 then
-    pn.penalty = 1000 -- supress orphan
+  local prev = getprev(curr)
+
+  if prev and prev.penalty then
+    -- repect user's penalty
   else
-    pn.penalty = 50
+    local pn = nodenew(penaltyid)
+    if not br then
+      pn.penalty = 10000
+    elseif type(brb) == "number" then
+      pn.penalty = brb
+    elseif par and nodecount(glyphid, curr) <= 2 then
+      pn.penalty = 1000 -- supress orphan
+    else
+      pn.penalty = 50
+    end
+    head = insert_before(head, curr, pn)
   end
+  if not fid then return head end -- penalty only for non-glyph box + cjk
 
   dim = dim or 0
   local gl = nodenew(glueid)
@@ -876,10 +895,8 @@ local function insert_glue_before (head, curr, par, br, brb, classic, ict, dim,
     setglue(gl, dim, str, str*0.6)
   end
 
-  local prev = getprev(curr)
   gl.attr = prev and prev.attr or curr.attr -- for less tagging
 
-  head = insert_before(head, curr, pn)
   return insert_before(head, curr, gl)
 end
 
@@ -894,35 +911,78 @@ local function maybe_linebreak (head, curr, pc, pcl, cc, old, fid, par)
       head = insert_glue_before(head, curr, par, br, brb, old, ict, dim, fid)
     end
   end
-  return head, cc, ccl
+  return head, cc, ccl, fid
+end
+
+local function process_cjk_punctuation_spacing (head, par)
+  local pcl, pc, pf, old = 0, false, false
+  local curr = head
+  while curr do
+    local id = curr.id
+    if id == glyphid and curr.lang ~= nohyphen then
+      local cc = curr.char
+      old = has_attribute(curr, classicattr)
+      local ccl = get_char_class(cc, old)
+      if old and intercharclass[pcl][ccl] then
+        local cf = charclass[cc] == 0 and pf or curr.font
+        head = maybe_linebreak(head, curr, pc, pcl, cc, old, cf, par)
+      end
+      pcl, pc, pf = ccl, cc, curr.font
+    elseif is_blocking_node(curr) then
+      if id == glueid and (curr.subtype >= spaceskip and curr.subtype <= parfillskip
+        or has_attribute(curr, inhibitglueattr)) then
+        pcl, pc, pf = 0, false, false
+      else
+        if pf and old and intercharclass[pcl][0] then
+          head = maybe_linebreak(head, curr, pc, pcl, 0x4E00, old, pf, par)
+        end
+        pcl, pc, pf = 0, 0x4E00, false
+      end
+    end
+    curr = getnext(curr)
+  end
+  return head
 end
 
 local function process_linebreak (head, par)
-  local curr, pc, pcl = head, false, 0
+  local curr, pc, pcl, pf = head, false, 0, false
   while curr do
     local id = curr.id
     if id == glyphid then
       local c = has_attribute(curr, unicodeattr) or curr.char
       if c and not is_combining(curr.char) then -- we are in pre-shaping stage
         local old = has_attribute(curr, classicattr)
-        head, pc, pcl = maybe_linebreak(head, curr, pc, pcl, c, old, curr.font, par)
+        local cjk = is_cjk_char(c)
+        local f = cjk and curr.font or pf or curr.font
+        if old and cjk and pc == -1 then -- penalty only (f is nil below) for non-glyph box + cjk
+          head = insert_glue_before(head, curr, par, true, false)
+          pc, pf, pcl = c, f, get_char_class(c, old)
+        else
+          head, pc, pcl, pf = maybe_linebreak(head, curr, pc, pcl, c, old, f, par)
+        end
       end
 
-    elseif id == hlistid and curr.list then
+    elseif (id == hlistid or id == vlistid) and is_blocking_node(curr) then
       local old = has_attribute(curr, classicattr)
       local c, f = hbox_char_font(curr, true)
-      if c and f then
-        head = maybe_linebreak(head, curr, pc, pcl, c, old, f, par)
+      if c then
+        head = maybe_linebreak(head, curr, pc, pcl, c, old, pf or f, par)
+      elseif old and pc and is_cjk_char(pc) then -- penalty only
+        head = insert_glue_before(head, curr, par, true, false)
       end
-      pc = hbox_char_font(curr)
-      pcl = pc and get_char_class(pc, old) or 0
+      c, f = hbox_char_font(curr)
+      pc, pf, pcl  = c or -1, pf or f, c and get_char_class(c, old) or 0
 
     elseif id == whatsitid and curr.mode == directmode then
       local glyf, c, fin = get_actualtext(curr)
       if c and fin and glyf then
         local old = has_attribute(glyf, classicattr)
-        head = maybe_linebreak(head, curr, pc, pcl, c, old, glyf.font, par)
-        pc, pcl, curr = fin, 0, goto_end_actualtext(curr)
+        if old and pc == -1 then -- penalty only
+          head = insert_glue_before(head, curr, par, true, false)
+        else
+          head = maybe_linebreak(head, curr, pc, pcl, c, old, glyf.font, par)
+        end
+        pc, pcl, curr, pf = fin, 0, goto_end_actualtext(curr), glyf.font
       end
 
     elseif id == mathid then
@@ -949,41 +1009,41 @@ local function do_interhangul_option (head, curr, pc, c, fontid, par)
     end
   end
 
-  return head, cc
+  return head, cc, fontid
 end
 
 local function process_interhangul (head, par)
-  local curr, pc = head, 0
+  local curr, pc, pf = head, 0, false
   while curr do
     local id = curr.id
     if id == glyphid then
       local c = has_attribute(curr, unicodeattr) or curr.char
       if c and not is_combining(curr.char) then -- we are in pre-shaping stage
-        head, pc = do_interhangul_option(head, curr, pc, c, curr.font, par)
+        head, pc, pf = do_interhangul_option(head, curr, pc, c, curr.font, par)
 
         if is_jungsong(c) or is_jongsong(c) or hangul_tonemark[c] then
           pc = 1
         end
       end
 
-    elseif id == hlistid and curr.list then
+    elseif id == hlistid and is_blocking_node(curr) then
       local c, f = hbox_char_font(curr, true)
-      if c and f then
-        head = do_interhangul_option(head, curr, pc, c, f, par)
+      if c then
+        head, pc, pf = do_interhangul_option(head, curr, pc, c, pf or f, par)
       end
-      c = hbox_char_font(curr)
-      pc = c and is_hangul_jamo(c) and 1 or 0
+      c, f = hbox_char_font(curr)
+      pc, pf = c and (is_hangul_jamo(c) or hangul_tonemark[c]) and 1 or 0, pf or f
 
     elseif id == whatsitid and curr.mode == directmode then
       local glyf, c = get_actualtext(curr)
       if c and glyf then
         head = do_interhangul_option(head, curr, pc, c, glyf.font, par)
-        pc, curr = 1, goto_end_actualtext(curr)
+        pc, curr, pf = 1, goto_end_actualtext(curr), glyf.font
       end
 
     elseif id == mathid then
       pc, curr = 0, end_of_math(curr)
-    elseif is_blocking_node(curr) or id == dirid then
+    elseif is_blocking_node(curr) then
       pc = 0
     end
     curr = getnext(curr)
@@ -995,13 +1055,15 @@ local function do_interlatincjk_option (head, curr, pc, pf, pcl, c, cf, par)
   local cc = is_cjk_char(c) and 1 or is_noncjk_char(c) and 2 or 0
   local old = has_attribute(curr, classicattr)
   local ccl = get_char_class(c, old)
+  local f = cc == 1 and cf or pf
 
   if cc*pc == 2 and curr.lang ~= nohyphen then
     local brb = cc == 2 or breakable_before[c] -- numletter != br_before
     if brb then
-      local f = cc == 1 and cf or pf
-      local dim = fontoptions.interlatincjk[f]
-      if dim then
+      local dimc = fontoptions.interlatincjk[cf] or 0
+      local dimp = fontoptions.interlatincjk[pf] or 0
+      local dim  = mathmax(dimc, dimp)
+      if dim ~= 0 then
         local ict = old and intercharclass[pcl][ccl] -- under classic env. only
         if ict then
           dim = fontoptions.intercharacter[f] or 0
@@ -1011,11 +1073,20 @@ local function do_interlatincjk_option (head, curr, pc, pf, pcl, c, cf, par)
     end
   end
 
-  return head, cc, cf, ccl
+  return head, cc, f, ccl
 end
 
 local function process_interlatincjk (head, par)
-  local curr, pc, pf, pcl = head, 0, 0, 0
+  local curr, pc, pf, pcl = head, 0, false, 0
+  while curr do
+    if curr.id == glyphid and is_cjk_char(curr.char) then
+      pf = curr.font
+      break
+    end
+    curr = getnext(curr)
+  end
+
+  curr = head
   while curr do
     local id = curr.id
     if id == glyphid then
@@ -1025,19 +1096,15 @@ local function process_interlatincjk (head, par)
         pc = breakable_after[c] and pc or 0
       end
 
-    elseif id == hlistid and curr.list then
+    elseif id == hlistid and is_blocking_node(curr) then
       local c, f = hbox_char_font(curr, true)
-      if c and f then
-        head = do_interlatincjk_option(head, curr, pc, pf, pcl, c, f, par)
+      if c then
+        head = do_interlatincjk_option(head, curr, pc, pf, pcl, c, pf or f, par)
       end
       c, f = hbox_char_font(curr)
-      if c and breakable_after[c] then
-        pc = is_cjk_char(c) and 1 or is_noncjk_char(c) and 2 or 0
-      else
-        pc = 0
-      end
+      pc  = c and breakable_after[c] and (is_cjk_char(c) and 1 or is_noncjk_char(c) and 2) or 0
       pcl = c and get_char_class(c, has_attribute(curr, classicattr)) or 0
-      pf  = f or 0
+      pf  = pf or f
 
     elseif id == whatsitid and curr.mode == directmode then
       local glyf, c = get_actualtext(curr)
@@ -1050,18 +1117,19 @@ local function process_interlatincjk (head, par)
       if pc == 1 then
         head = do_interlatincjk_option(head, curr, pc, pf, pcl, 0x30, pf, par)
       end
-      pc, pf, pcl = 2, 0, 0
+      pc, pcl = 2, 0
       curr = end_of_math(curr)
 
     elseif id == dirid then
       if pc == 1 and curr.dir:sub(1,1) == "+" then
         head = do_interlatincjk_option(head, curr, pc, pf, pcl, 0x30, pf, par)
-        pc, pf, pcl = 0, 0, 0
       end
+      pc, pcl = 0, 0
 
     elseif is_blocking_node(curr) then
-      pc, pf, pcl = 0, 0, 0
+      pc, pcl = 0, 0
     end
+
     curr = getnext(curr)
   end
   return head
@@ -1117,46 +1185,27 @@ end
 -- remove classic spaces
 
 local function process_remove_spaces (head)
-  local curr, to_free = head, {}
+  local curr, to_free, sp, cjk = head, {}, false, false
   while curr do
     local id = curr.id
-    if id == glueid then
-      if curr.subtype == spaceskip and has_attribute(curr, classicattr) then
-
-        for k, v in pairs{ p = getprev(curr), n = getnext(curr) } do
-          local ok
-          while v do
-            local id = v.id
-            if id ~= whatsitid -- skip whatsit or kern except userkern
-              and ( id ~= kernid or v.subtype == userkern ) then
-
-              local vchar, vfont
-              if id == glyphid and v.lang ~= nohyphen then
-                local c = has_attribute(v, unicodeattr) or v.char or 0
-                if is_combining(c) then
-                  v = getprev(v) or v
-                end
-                vchar, vfont = has_attribute(v, unicodeattr) or v.char, v.font
-              elseif id == hlistid and v.list then
-                vchar, vfont = hbox_char_font(v, k == "n")
-              end
-              if vchar and vfont and fontoptions.removeclassicspaces[vfont] then
-                ok = is_cjk_char(vchar)
-              end
-
-              break
-            end
-            v = k == "p" and getprev(v) or getnext(v)
-          end
-          if ok then
-            head = noderemove(head, curr)
-            tableinsert(to_free, curr)
-            break
-          end
+    if id == glyphid and curr.lang ~= nohyphen and has_attribute(curr, classicattr)
+      and fontoptions.removeclassicspaces[curr.font] then
+      local c = has_attribute(curr, unicodeattr) or curr.char or 0
+      if not is_combining(c) then
+        local cc = is_cjk_char(c) and 1 or 0
+        if sp and cjk and (cjk + cc) == 2 then
+          head = noderemove(head, sp)
+          tableinsert(to_free, sp)
         end
+        cjk, sp = cc, false
       end
+    elseif cjk and id == glueid
+      and curr.subtype == spaceskip and has_attribute(curr, classicattr) then
+      sp = curr
     elseif id == mathid then
-      curr = end_of_math(curr)
+      curr, cjk, sp = end_of_math(curr), false, false
+    elseif id == hlistid or is_blocking_node(curr) then
+      cjk, sp = false, false
     end
     curr = getnext(curr)
   end
@@ -1394,7 +1443,7 @@ local function process_dotemph (head)
         if init then
           local basewd = curr.width
           if basewd >= fontoptions.en_size[curr.font] then
-            local box = nodecopy(dotemphbox[dotattr]).list
+            local box = nodecopy(dotemphbox[dotattr])
             -- bypass unwanted nodes injected by some other packages
             while box.id ~= hlistid do
               warning[[\dotemph should be an hbox]]
@@ -1428,7 +1477,7 @@ local function process_dotemph (head)
             if shift ~= 0 then
               local list = box.list
               local k = nodenew(kernid)
-              k.kern = shift
+              k.kern, k.subtype = shift, userkern
               box.list = insert_before(list, list, k)
             end
 
@@ -1437,6 +1486,10 @@ local function process_dotemph (head)
 
             box.width = 0
             head, curr = insert_after(head, curr, box)
+
+            local k = nodenew(kernid)
+            k.subtype, k.kern = userkern, 0
+            head = insert_before(head, curr, k)
           end
         end
       end
@@ -1536,6 +1589,7 @@ local function draw_uline (head, curr, parent, t, final)
       g.attr = curr.attr
       local k = nodenew(kernid)
       k.kern = -len
+      k.subtype = userkern
       head = insert_before(head, start, g)
       head = insert_before(head, start, k)
     end
@@ -1657,7 +1711,7 @@ local function process_ruby_post_linebreak (head)
           if side ~= 0 then
             local list = ruby.list
             local k = nodenew(kernid)
-            k.kern = side
+            k.kern, k.subtype = side, userkern
             ruby.list = insert_before(list, list, k)
           end
           ruby.width = 0
@@ -1673,6 +1727,10 @@ local function process_ruby_post_linebreak (head)
 
           ruby.shift = shift - ascender - descender - ruby_t[2] -- rubysep
           head = insert_before(head, curr, ruby)
+
+          local k = nodenew(kernid)
+          k.subtype, k.kern = userkern, 0
+          head = insert_before(head, curr, k) -- prevent possible glue insertion (unhbox etc)
         end
         rubybox[rubyid] = nil
       end
@@ -1791,129 +1849,132 @@ end
 
 -- vertical font
 
-local streamreader = utilities.files
-local openfile     = streamreader.open
-local closefile    = streamreader.close
-local readstring   = streamreader.readstring
-local readulong    = streamreader.readcardinal4
-local readushort   = streamreader.readcardinal2
-local readfixed    = streamreader.readfixed4
-local readshort    = streamreader.readinteger2
-local setpos       = streamreader.setposition
+local get_tsb_table
+do
+  local streamreader = utilities.files
+  local openfile     = streamreader.open
+  local closefile    = streamreader.close
+  local readstring   = streamreader.readstring
+  local readulong    = streamreader.readcardinal4
+  local readushort   = streamreader.readcardinal2
+  local readfixed    = streamreader.readfixed4
+  local readshort    = streamreader.readinteger2
+  local setpos       = streamreader.setposition
 
-local function get_otf_tables (f, subfont)
-  if f then
-    local sfntversion = readstring(f,4)
-    if sfntversion == "ttcf" then
-      local ttcversion = readfixed(f)
-      local numfonts   = readulong(f)
-      if subfont >= 1 and subfont <= numfonts then
-        local offsets = {}
-        for i = 1, numfonts do
-          offsets[i] = readulong(f)
+  local function get_otf_tables (f, subfont)
+    if f then
+      local sfntversion = readstring(f,4)
+      if sfntversion == "ttcf" then
+        local ttcversion = readfixed(f)
+        local numfonts   = readulong(f)
+        if subfont >= 1 and subfont <= numfonts then
+          local offsets = {}
+          for i = 1, numfonts do
+            offsets[i] = readulong(f)
+          end
+          setpos(f, offsets[subfont])
+          sfntversion = readstring(f,4)
         end
-        setpos(f, offsets[subfont])
-        sfntversion = readstring(f,4)
+      end
+      if sfntversion == "OTTO" or sfntversion == "true" or sfntversion == "\0\1\0\0" then
+        local numtables     = readushort(f)
+        local searchrange   = readushort(f)
+        local entryselector = readushort(f)
+        local rangeshift    = readushort(f)
+        local tables        = {}
+        for i= 1, numtables do
+          local tag = readstring(f,4)
+          tables[tag] = {
+            checksum = readulong(f),
+            offset   = readulong(f),
+            length   = readulong(f),
+          }
+        end
+        return tables
       end
     end
-    if sfntversion == "OTTO" or sfntversion == "true" or sfntversion == "\0\1\0\0" then
-      local numtables     = readushort(f)
-      local searchrange   = readushort(f)
-      local entryselector = readushort(f)
-      local rangeshift    = readushort(f)
-      local tables        = {}
-      for i= 1, numtables do
-        local tag = readstring(f,4)
-        tables[tag] = {
-          checksum = readulong(f),
-          offset   = readulong(f),
-          length   = readulong(f),
+  end
+
+  local function read_maxp (f, t)
+    if f and t then
+      setpos(f, t.offset)
+      return {
+        version   = readfixed(f),
+        numglyphs = readushort(f),
+      }
+    end
+  end
+
+  local function read_vhea (f, t)
+    if f and t then
+      setpos(f, t.offset)
+      return {
+        version               = readfixed(f),
+        ascent                = readshort(f),
+        descent               = readshort(f),
+        lineGap               = readshort(f),
+        advanceheightmax      = readshort(f),
+        mintopsidebearing     = readshort(f),
+        minbottomsidebrearing = readshort(f),
+        ymaxextent            = readshort(f),
+        caretsloperise        = readshort(f),
+        caretsloperun         = readshort(f),
+        caretoffset           = readshort(f),
+        reserved1             = readshort(f),
+        reserved2             = readshort(f),
+        reserved3             = readshort(f),
+        reserved4             = readshort(f),
+        metricdataformat      = readshort(f),
+        numheights            = readushort(f),
+      }
+    end
+  end
+
+  local function read_vmtx (f, t, numofheights, numofglyphs)
+    if f and t and numofheights and numofglyphs then
+      setpos(f, t.offset)
+      local vmtx = {}
+      local height = 0
+      for i = 0, numofheights-1 do
+        height = readushort(f)
+        vmtx[i] = {
+          ht  = height,
+          tsb = readshort(f),
         }
       end
-      return tables
+      for i = numofheights, numofglyphs-1 do
+        vmtx[i] = {
+          ht  = height,
+          tsb = readshort(f),
+        }
+      end
+      return vmtx
     end
   end
-end
 
-local function read_maxp (f, t)
-  if f and t then
-    setpos(f, t.offset)
-    return {
-      version   = readfixed(f),
-      numglyphs = readushort(f),
-    }
-  end
-end
+  local tsb_font_data = {}
 
-local function read_vhea (f, t)
-  if f and t then
-    setpos(f, t.offset)
-    return {
-      version               = readfixed(f),
-      ascent                = readshort(f),
-      descent               = readshort(f),
-      lineGap               = readshort(f),
-      advanceheightmax      = readshort(f),
-      mintopsidebearing     = readshort(f),
-      minbottomsidebrearing = readshort(f),
-      ymaxextent            = readshort(f),
-      caretsloperise        = readshort(f),
-      caretsloperun         = readshort(f),
-      caretoffset           = readshort(f),
-      reserved1             = readshort(f),
-      reserved2             = readshort(f),
-      reserved3             = readshort(f),
-      reserved4             = readshort(f),
-      metricdataformat      = readshort(f),
-      numheights            = readushort(f),
-    }
-  end
-end
-
-local function read_vmtx (f, t, numofheights, numofglyphs)
-  if f and t and numofheights and numofglyphs then
-    setpos(f, t.offset)
-    local vmtx = {}
-    local height = 0
-    for i = 0, numofheights-1 do
-      height = readushort(f)
-      vmtx[i] = {
-        ht  = height,
-        tsb = readshort(f),
-      }
+  function get_tsb_table (filename, subfont)
+    subfont = tonumber(subfont) or 1
+    local key = stringformat("%s::%s", filename, subfont)
+    if tsb_font_data[key] then
+      return tsb_font_data[key]
     end
-    for i = numofheights, numofglyphs-1 do
-      vmtx[i] = {
-        ht  = height,
-        tsb = readshort(f),
-      }
+    local f = openfile(filename, true) -- true: zero-based
+    if f then
+      local vmtx
+      local tables = get_otf_tables(f, subfont)
+      if tables then
+        local vhea = read_vhea(f, tables.vhea)
+        local numofheights = vhea and vhea.numheights
+        local maxp = read_maxp(f, tables.maxp)
+        local numofglyphs = maxp and maxp.numglyphs
+        vmtx = read_vmtx(f, tables.vmtx, numofheights, numofglyphs)
+      end
+      closefile(f)
+      tsb_font_data[key] = vmtx
+      return vmtx
     end
-    return vmtx
-  end
-end
-
-local tsb_font_data = {}
-
-local function get_tsb_table (filename, subfont)
-  subfont = tonumber(subfont) or 1
-  local key = stringformat("%s::%s", filename, subfont)
-  if tsb_font_data[key] then
-    return tsb_font_data[key]
-  end
-  local f = openfile(filename, true) -- true: zero-based
-  if f then
-    local vmtx
-    local tables = get_otf_tables(f, subfont)
-    if tables then
-      local vhea = read_vhea(f, tables.vhea)
-      local numofheights = vhea and vhea.numheights
-      local maxp = read_maxp(f, tables.maxp)
-      local numofglyphs = maxp and maxp.numglyphs
-      vmtx = read_vmtx(f, tables.vmtx, numofheights, numofglyphs)
-    end
-    closefile(f)
-    tsb_font_data[key] = vmtx
-    return vmtx
   end
 end
 
@@ -2239,6 +2300,7 @@ add_to_callback("pre_shaping_filter", function(h, gc)
            or gc == "insert"
            or gc == "vcenter"
   h = call_callback("luatexko_prelinebreak_first", h, par)
+  h = process_cjk_punctuation_spacing(h, par)
   h = call_callback("luatexko_prelinebreak_second", h, par)
   return process_linebreak(h, par)
 end, "luatexko.pre_shaping_filter")
