@@ -1,5 +1,5 @@
 --% Kale Ewasiuk (kalekje@gmail.com)
---% 2025-01-11
+--% 2025-02-14
 --% Copyright (C) 2021-2025 Kale Ewasiuk
 --%
 --% Permission is hereby granted, free of charge, to any person obtaining a copy
@@ -902,7 +902,6 @@ function penlight.findfiles(kv)
     local files = penlight.List{}
     for fn, ext in penlight.seq.prod(kv.fn, kv.ext) do
         for _, dir in ipairs(kv.dir) do
-            penlight.wrth(dir..fn..ext)
             files:extend(getfiles(dir, fn..ext))
         end
     end
@@ -911,6 +910,7 @@ function penlight.findfiles(kv)
     return files
 end
 
+penlight.dir.findfiles = penlight.findfiles
 
 --todo add doc
 function penlight.utils.filterfiles(...)
@@ -1145,7 +1145,7 @@ penlight.tex.writePDFmetadata = function(t) -- write PDF metadata to xmpdata fil
   for k, v in pairs(t) do
     str = str..'\\'..k..'{'..v..'}'..'\n'
   end
-  penlight.utils.writefile(tex.jobname..'.xmpdata', str)
+  penlight.utils.writefile(tex.jobname:gsub('"','')..'.xmpdata', str)
 end
 
 
