@@ -1,13 +1,13 @@
-# $Id: TLPDB.pm 63068 2022-04-18 05:58:07Z preining $
+# $Id: TLPDB.pm 71592 2024-06-22 21:01:54Z karl $
 # TeXLive::TLPDB.pm - tlpdb plain text database files.
-# Copyright 2007-2021 Norbert Preining
+# Copyright 2007-2024 Norbert Preining
 # This file is licensed under the GNU General Public License version 2
 # or any later version.
 
 use strict; use warnings;
 package TeXLive::TLPDB;
 
-my $svnrev = '$Revision: 63068 $';
+my $svnrev = '$Revision: 71592 $';
 my $_modulerevision = ($svnrev =~ m/: ([0-9]+) /) ? $1 : "unknown";
 sub module_revision { return $_modulerevision; }
 
@@ -98,7 +98,7 @@ use TeXLive::TLConfig qw($CategoriesRegexp $DefaultCategory $InfraLocation
       $RelocPrefix $RelocTree);
 use TeXLive::TLCrypto;
 use TeXLive::TLPOBJ;
-use TeXLive::TLUtils qw(dirname mkdirhier member win32 info log debug ddebug
+use TeXLive::TLUtils qw(dirname mkdirhier member wndws info log debug ddebug
                         tlwarn basename download_file merge_into tldie
                         system_pipe);
 use TeXLive::TLWinGoo;
@@ -362,9 +362,24 @@ sub from_file {
         # do nothing
       } else {
         unlink($tlpdbfile);
-        tldie(  "$0: TLPDB::from_file could not initialize from: $path\n"
-              . "$0: Maybe the repository setting should be changed.\n"
-              . "$0: More info: https://tug.org/texlive/acquire.html\n");
+        my $diemsg = <<END_DOWNLOAD_FAILURE_MSG;
+$0: TLPDB::from_file could not get texlive.tlpdb from: $path
+Maybe the repository setting should be changed.
+More info: https://tug.org/texlive/acquire.html
+END_DOWNLOAD_FAILURE_MSG
+
+        # If they have the Cygwin wget.exe, some other problem.
+        if ($^O eq 'cygwin'
+            && (! -x "/usr/bin/curl.exe" && ! -x "/usr/bin/wget.exe")) {
+          $diemsg .= <<END_CYGWIN_WGET_MSG;
+
+It seems you are using Cygwin and have not installed Cygwin's
+curl or wget. See the TeX Live Guide information on Cygwin for required
+and recommended packages:
+  https://tug.org/texlive/doc/texlive-en/texlive-en.html#cygwin
+END_CYGWIN_WGET_MSG
+        }
+        tldie($diemsg);
       }
     }
     # if we are still here, then either the xz version was downloaded
@@ -875,10 +890,10 @@ sub expand_dependencies {
             $install{"$foo.$a"} = $install{$foo}
               if defined($self->get_package("$foo.$a"));
           }
-        } elsif ($p_dep =~ m/^(.*)\.win32$/) {
-          # a win32 package should *only* be installed if we are installing
-          # the win32 arch
-          if (grep(/^win32$/,@archs)) {
+        } elsif ($p_dep =~ m/^(.*)\.windows$/) {
+          # a windows package should *only* be installed if we are installing
+          # the windows arch
+          if (grep(/^windows$/,@archs)) {
             $install{$p_dep} = 0;
           }
         } else {
@@ -1187,8 +1202,10 @@ sub platform {
   # try to deduce the platform
   my $self = shift;
   my $ret = $self->setting("platform");
+  #print STDERR "platform $ret set in tlpdb\n" if defined $ret;
   return $ret if defined $ret;
   # the platform setting wasn't found in the tlpdb, try TLUtils::platform
+  #print STDERR "Setting platform to ",TeXLive::TLUtils::platform(), "\n";
   return TeXLive::TLUtils::platform();
 }
 
@@ -1656,42 +1673,11 @@ sub install_package_files {
     # place from where files should be installed
     if (!_install_data ($tmpdir, \@installfiles, $reloc, \@installfiles,
                         $self)) {
-      tlwarn("TLPDB::install_package_files: couldn't install_data files: "
+      tlwarn("TLPDB::install_package_files: couldn't _install_data files: "
              . "@installfiles\n"); 
       next;
     }
-    if ($reloc) {
-      if ($self->setting("usertree")) {
-        $tlpobj->cancel_reloc_prefix;
-      } else {
-        $tlpobj->replace_reloc_prefix;
-      }
-      $tlpobj->relocated(0);
-    }
-    my $tlpod = $self->root . "/tlpkg/tlpobj";
-    mkdirhier( $tlpod );
-    open(TMP,">$tlpod/".$tlpobj->name.".tlpobj") or
-      die("Cannot open tlpobj file for ".$tlpobj->name);
-    $tlpobj->writeout(\*TMP);
-    close(TMP);
-    $self->add_tlpobj($tlpobj);
-    $self->save;
-    TeXLive::TLUtils::announce_execute_actions("enable", $tlpobj);
-    # do the postinstallation actions
-    #
-    # Run the post installation code in the postaction tlpsrc entries
-    # in case we are on w32 and the admin did install for himself only
-    # we switch off admin mode
-    if (win32() && admin() && !$self->option("w32_multi_user")) {
-      non_admin();
-    }
-    # for now desktop_integration maps to both installation
-    # of desktop shortcuts and menu items, but we can split them later
-    &TeXLive::TLUtils::do_postaction("install", $tlpobj,
-      $self->option("file_assocs"),
-      $self->option("desktop_integration"),
-      $self->option("desktop_integration"),
-      $self->option("post_code"));
+    _post_install_pkg ($self, $tlpobj);
 
     # remember that we installed this package correctly
     $ret++;
@@ -1699,12 +1685,14 @@ sub install_package_files {
   return $ret;
 }
 
-
+
 =pod
 
 =item C<< $tlpdb->install_package($pkg, $dest_tlpdb [, $tag]) >>
 
-Installs the package $pkg into $dest_tlpdb.
+Installs the package $pkg into $dest_tlpdb. Returns a reference to the
+package, or undef if failure.
+
 If C<$tag> is present and the tlpdb is virtual, tries to install $pkg
 from the repository tagged with $tag.
 
@@ -1719,7 +1707,7 @@ sub install_package {
       } else {
         tlwarn("TLPDB::install_package: package $pkg not found"
                . " in repository $tag\n");
-        return;
+        return undef;
       }
     } else {
       my ($maxtag, $maxrev, $maxtlp, $maxtlpdb)
@@ -1733,7 +1721,7 @@ sub install_package {
     }
     return $self->not_virtual_install_package($pkg, $totlpdb);
   }
-  return;
+  return undef;
 }
 
 sub not_virtual_install_package {
@@ -1831,7 +1819,7 @@ sub not_virtual_install_package {
       # will never be installed, and we do already check that there
       # are at all src/doc files, which in split packages of the form 
       # foo.ARCH are not present. And if they are present, than that is fine,
-      # too (bin-foobar.win32.doc.tar.xz)
+      # too (bin-foobar.windows.doc.tar.xz)
       # - there are actually src/doc files present
       if ($container_src_split && $opt_src && $tlpobj->srcfiles) {
         my $srccontainer = $container;
@@ -1880,55 +1868,73 @@ sub not_virtual_install_package {
     if (!$real_opt_doc) {
       $tlpobj->clear_docfiles;
     }
-    # if a package is relocatable we have to cancel the reloc prefix
-    # and unset the relocated setting
-    # before we save it to the local tlpdb
-    if ($tlpobj->relocated) {
-      if ($totlpdb->setting("usertree")) {
-        $tlpobj->cancel_reloc_prefix;
-      } else {
-        $tlpobj->replace_reloc_prefix;
-      }
-      $tlpobj->relocated(0);
-    }
-    # we have to write out the tlpobj file since it is contained in the
-    # archives (.tar.xz) but at DVD install time we don't have them
-    my $tlpod = $totlpdb->root . "/tlpkg/tlpobj";
-    mkdirhier($tlpod);
-    my $count = 0;
-    my $tlpobj_file = ">$tlpod/" . $tlpobj->name . ".tlpobj";
-    until (open(TMP, $tlpobj_file)) {
-      # The open might fail for no good reason on Windows.
-      # Try again for a while, but not forever.
-      if ($count++ == 100) { die "$0: open($tlpobj_file) failed: $!"; }
-      select (undef, undef, undef, .1);  # sleep briefly
-    }
-    $tlpobj->writeout(\*TMP);
-    close(TMP);
-    $totlpdb->add_tlpobj($tlpobj);
-    $totlpdb->save;
-    # compute the return value
-    TeXLive::TLUtils::announce_execute_actions("enable", $tlpobj);
-    # do the postinstallation actions
-    #
-    # Run the post installation code in the postaction tlpsrc entries
-    # in case we are on w32 and the admin did install for himself only
-    # we switch off admin mode
-    if (win32() && admin() && !$totlpdb->option("w32_multi_user")) {
-      non_admin();
-    }
-    # for now desktop_integration maps to both installation
-    # of desktop shortcuts and menu items, but we can split them later
-    &TeXLive::TLUtils::do_postaction("install", $tlpobj,
-      $totlpdb->option("file_assocs"),
-      $totlpdb->option("desktop_integration"),
-      $totlpdb->option("desktop_integration"),
-      $totlpdb->option("post_code"));
+    _post_install_pkg ($totlpdb, $tlpobj);
   }
   return 1;
 }
 
-#
+# In TLPDB, Do post-install stuff for TLPOBJ:
+# - cancel relocation stuff
+# - write the tlpobj
+# - handle post-installation actions
+# 
+sub _post_install_pkg {
+  my ($tlpdb,$tlpobj) = @_;
+  
+  # if a package is relocatable we have to cancel the reloc prefix
+  # and unset the relocated setting
+  # before we save it to the local tlpdb
+  if ($tlpobj->relocated) {
+    if ($tlpdb->setting("usertree")) {
+      $tlpobj->cancel_reloc_prefix;
+    } else {
+      $tlpobj->replace_reloc_prefix;
+    }
+    $tlpobj->relocated(0);
+  }
+  # we have to write out the tlpobj file since it is contained in the
+  # archives (.tar.xz) but at DVD install time we don't have them
+  my $tlpod = $tlpdb->root . "/tlpkg/tlpobj";
+  mkdirhier($tlpod);
+  my $count = 0;
+  my $tlpobj_file = ">$tlpod/" . $tlpobj->name . ".tlpobj";
+  until (open(TMP, $tlpobj_file)) {
+    # The open might fail for no good reason on Windows.
+    # Try again for a while, but not forever.
+    if ($count++ == 100) { die "$0: open($tlpobj_file) failed: $!"; }
+    select(undef, undef, undef, .1);  # sleep briefly
+  }
+  $tlpobj->writeout(\*TMP);
+  close(TMP);
+  $tlpdb->add_tlpobj($tlpobj);
+  $tlpdb->save;
+  #
+  # do postinstallation actions.
+  #
+  # Remember to do any postactions, including recording whether files
+  # have changed.
+  TeXLive::TLUtils::announce_execute_actions("enable", $tlpobj);
+  #
+  # If this was context, remember to do its cache.
+  if ($tlpobj->name eq "context") {
+    TeXLive::TLUtils::announce_execute_actions("context-cache", $tlpobj);
+  }
+  #  
+  # Run the post installation code in the postaction tlpsrc entries
+  # in case we are on w32 and the admin did install for himself only
+  # we switch off admin mode
+  if (wndws() && admin() && !$tlpdb->option("w32_multi_user")) {
+    non_admin();
+  }
+  # for now desktop_integration maps to both installation
+  # of desktop shortcuts and menu items, but we can split them if need be.
+  &TeXLive::TLUtils::do_postaction("install", $tlpobj,
+    $tlpdb->option("file_assocs"),
+    $tlpdb->option("desktop_integration"),
+    $tlpdb->option("desktop_integration"),
+    $tlpdb->option("post_code"));
+}
+
 # _install_data
 # actually does the installation work
 # returns 1 on success and 0 on error
@@ -1937,7 +1943,7 @@ sub not_virtual_install_package {
 # otherwise it is a tlpdb from where to install
 #
 sub _install_data {
-  my ($self, $what, $reloc, $filelistref, $totlpdb, $whatsize, $whatcheck) = @_;
+  my ($self, $what, $reloc, $filelistref, $totlpdb, $whatsize, $whatcheck) =@_;
 
   my $target = $totlpdb->root;
   my $tempdir = TeXLive::TLUtils::tl_tmpdir();
@@ -1995,6 +2001,7 @@ sub _install_data {
   }
 }
 
+
 =pod
 
 =item << $tlpdb->remove_package($pkg, %options) >>
@@ -2012,7 +2019,9 @@ sub remove_package {
   my $tlp = $localtlpdb->get_package($pkg);
   my $usertree = $localtlpdb->setting("usertree");
   if (!defined($tlp)) {
-    tlwarn ("TLPDB: package not present, so nothing to remove: $pkg\n");
+    # we should not be called.
+    tlwarn ("TLPDB::remove_package: package not present, ",
+            "so nothing to remove: $pkg\n");
   } else {
     my $currentarch = $self->platform();
     if ($pkg eq "texlive.infra" || $pkg eq "texlive.infra.$currentarch") {
@@ -2023,8 +2032,10 @@ sub remove_package {
     my $Master = $localtlpdb->root;
     chdir ($Master) || die "chdir($Master) failed: $!";
     my @files = $tlp->all_files;
+    #
     # also remove the .tlpobj file
     push @files, "tlpkg/tlpobj/$pkg.tlpobj";
+    #
     # and the ones from src/doc splitting
     if (-r "tlpkg/tlpobj/$pkg.source.tlpobj") {
       push @files, "tlpkg/tlpobj/$pkg.source.tlpobj";
@@ -2105,17 +2116,39 @@ sub remove_package {
         0, # tlpdbopt_desktop_integration, desktop part
         $localtlpdb->option("post_code"));
     }
-    # 
-    my @removals = &TeXLive::TLUtils::removed_dirs (@goodfiles);
+    # we want to check whether we can actually remove files
+    # there might be various reasons that this fails, like texmf-dist
+    # directory suddently becoming ro (for whatever definition of
+    # suddenly).
+    my (%by_dirs, %removed_dirs) = &TeXLive::TLUtils::all_dirs_and_removed_dirs (@goodfiles);
+    my @removals = keys %removed_dirs;
+
+    # we have already check for the existence of the dirs returned
+    for my $d (keys %by_dirs) {
+      if (! &TeXLive::TLUtils::dir_writable($d)) {
+        tlwarn("TLPDB::remove_package: directories are not writable, cannot remove files: $d\n");
+        return 0;
+      }
+    }
+
     # now do the removal
     for my $entry (@goodfiles) {
-      unlink $entry;
+      # sometimes the files might not be there: 1) we remove .tlpobj
+      # explicitly above; 2) we're called from tl-update-containers
+      # to update the network tlpdb, and that doesn't have an expanded
+      # texmf-dist.
+      next unless -e $entry;
+      #
+      unlink($entry)
+      || tlwarn("TLPDB::remove_package: Could not unlink $entry: $!\n");
     }
     for my $d (@removals) {
-      rmdir $d;
+      rmdir($d)
+      || tlwarn("TLPDB::remove_package: Could not rmdir $d: $!\n")
     }
     $localtlpdb->remove_tlpobj($pkg);
     TeXLive::TLUtils::announce_execute_actions("disable", $tlp);
+    
     # should we save at each removal???
     # advantage: the tlpdb actually reflects what is installed
     # disadvantage: removing a collection calls the save routine several times
@@ -2125,7 +2158,7 @@ sub remove_package {
     # Run the post installation code in the postaction tlpsrc entries
     # in case we are on w32 and the admin did install for himself only
     # we switch off admin mode
-    if (win32() && admin() && !$localtlpdb->option("w32_multi_user")) {
+    if (wndws() && admin() && !$localtlpdb->option("w32_multi_user")) {
       non_admin();
     }
     #
@@ -2134,6 +2167,7 @@ sub remove_package {
     # files are already removed.
     # Again, desktop integration maps to desktop and menu links
     if (!$opts{'nopostinstall'}) {
+      debug(" TLPDB::remove_package: running remove postinstall\n");
       &TeXLive::TLUtils::do_postaction("remove", $tlp,
         $localtlpdb->option("file_assocs"),
         $localtlpdb->option("desktop_integration"),
@@ -2529,7 +2563,7 @@ The purpose of virtual databases is to collect several data sources
 and present them in one way. The normal functions will always return
 the best candidate for the set of functions.
 
-More docs to be written someday, maybe.
+More docs to be written if there is any demand.
 
 =over 4
 
@@ -2998,4 +3032,4 @@ GNU General Public License Version 2 or later.
 ### tab-width: 2
 ### indent-tabs-mode: nil
 ### End:
-# vim:set tabstop=2 expandtab autoindent: #
+# vim:set tabstop=2 shiftwidth=2 expandtab autoindent: #
